@@ -16,18 +16,22 @@ import itertools
 
 import netaddr
 
-from neutron.db import api as db_api
-from neutron.db import common_db_mixin as common_db
 from neutron.db import l3_dvr_db
 from neutron.db.models import address_scope as address_scope_db
 from neutron.db.models import l3 as l3_db
 from neutron.db.models import l3_attrs as l3_attrs_db
 from neutron.db import models_v2
+from neutron.objects import ports
+from neutron.objects import subnet as subnet_obj
+from neutron.objects import subnetpool as subnetpool_obj
 from neutron.plugins.ml2 import models as ml2_models
 
 from neutron_lib.api import validators
 from neutron_lib import constants as lib_consts
+from neutron_lib.db import api as db_api
 from neutron_lib.db import model_base
+from neutron_lib.db import model_query
+from neutron_lib.db import utils as db_utils
 from neutron_lib import exceptions as n_exc
 from neutron_lib.exceptions import l3 as l3_exc
 from oslo_db import exception as oslo_db_exc
@@ -44,6 +48,7 @@ from neutron_dynamic_routing.extensions import bgp as bgp_ext
 
 DEVICE_OWNER_ROUTER_GW = lib_consts.DEVICE_OWNER_ROUTER_GW
 DEVICE_OWNER_ROUTER_INTF = lib_consts.DEVICE_OWNER_ROUTER_INTF
+DEVICE_OWNER_DVR_INTERFACE = lib_consts.DEVICE_OWNER_DVR_INTERFACE
 
 
 class BgpSpeakerPeerBinding(model_base.BASEV2):
@@ -123,7 +128,7 @@ class BgpPeer(model_base.BASEV2,
     password = sa.Column(sa.String(255), nullable=True)
 
 
-class BgpDbMixin(common_db.CommonDbMixin):
+class BgpDbMixin(object):
 
     def create_bgp_speaker(self, context, bgp_speaker):
         uuid = uuidutils.generate_uuid()
@@ -133,23 +138,22 @@ class BgpDbMixin(common_db.CommonDbMixin):
     def get_bgp_speakers(self, context, filters=None, fields=None,
                          sorts=None, limit=None, marker=None,
                          page_reverse=False):
-        with db_api.context_manager.reader.using(context):
-            return self._get_collection(context, BgpSpeaker,
-                                        self._make_bgp_speaker_dict,
-                                        filters=filters, fields=fields,
-                                        sorts=sorts, limit=limit,
-                                        page_reverse=page_reverse)
+        with db_api.CONTEXT_READER.using(context):
+            return model_query.get_collection(
+                context, BgpSpeaker, self._make_bgp_speaker_dict,
+                filters=filters, fields=fields, sorts=sorts, limit=limit,
+                page_reverse=page_reverse)
 
     def get_bgp_speaker(self, context, bgp_speaker_id, fields=None):
-        with db_api.context_manager.reader.using(context):
+        with db_api.CONTEXT_READER.using(context):
             bgp_speaker = self._get_bgp_speaker(context, bgp_speaker_id)
             return self._make_bgp_speaker_dict(bgp_speaker, fields)
 
     def get_bgp_speaker_with_advertised_routes(self, context,
                                                bgp_speaker_id):
         bgp_speaker_attrs = ['id', 'local_as', 'tenant_id']
-        bgp_peer_attrs = ['peer_ip', 'remote_as', 'password']
-        with db_api.context_manager.reader.using(context):
+        bgp_peer_attrs = ['peer_ip', 'remote_as', 'auth_type', 'password']
+        with db_api.CONTEXT_READER.using(context):
             bgp_speaker = self.get_bgp_speaker(context, bgp_speaker_id,
                                                fields=bgp_speaker_attrs)
             res = dict((k, bgp_speaker[k]) for k in bgp_speaker_attrs)
@@ -163,7 +167,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
 
     def update_bgp_speaker(self, context, bgp_speaker_id, bgp_speaker):
         bp = bgp_speaker[bgp_ext.BGP_SPEAKER_BODY_KEY_NAME]
-        with db_api.context_manager.writer.using(context):
+        with db_api.CONTEXT_WRITER.using(context):
             bgp_speaker_db = self._get_bgp_speaker(context, bgp_speaker_id)
             bgp_speaker_db.update(bp)
 
@@ -173,7 +177,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
     def _save_bgp_speaker(self, context, bgp_speaker, uuid):
         ri = bgp_speaker[bgp_ext.BGP_SPEAKER_BODY_KEY_NAME]
         ri['tenant_id'] = context.tenant_id
-        with db_api.context_manager.writer.using(context):
+        with db_api.CONTEXT_WRITER.using(context):
             res_keys = ['local_as', 'tenant_id', 'name', 'ip_version',
                         'advertise_floating_ip_host_routes',
                         'advertise_tenant_networks']
@@ -198,7 +202,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
 
     def add_gateway_network(self, context, bgp_speaker_id, network_info):
         network_id = self._get_id_for(network_info, 'network_id')
-        with db_api.context_manager.writer.using(context):
+        with db_api.CONTEXT_WRITER.using(context):
             try:
                 self._save_bgp_speaker_network_binding(context,
                                                        bgp_speaker_id,
@@ -210,7 +214,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
         return {'network_id': network_id}
 
     def remove_gateway_network(self, context, bgp_speaker_id, network_info):
-        with db_api.context_manager.writer.using(context):
+        with db_api.CONTEXT_WRITER.using(context):
             network_id = self._get_id_for(network_info, 'network_id')
             self._remove_bgp_speaker_network_binding(context,
                                                      bgp_speaker_id,
@@ -218,7 +222,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
         return {'network_id': network_id}
 
     def delete_bgp_speaker(self, context, bgp_speaker_id):
-        with db_api.context_manager.writer.using(context):
+        with db_api.CONTEXT_WRITER.using(context):
             bgp_speaker_db = self._get_bgp_speaker(context, bgp_speaker_id)
             context.session.delete(bgp_speaker_db)
 
@@ -229,7 +233,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
         if auth_type == 'md5' and not password:
             raise bgp_ext.InvalidBgpPeerMd5Authentication()
 
-        with db_api.context_manager.writer.using(context):
+        with db_api.CONTEXT_WRITER.using(context):
             res_keys = ['tenant_id', 'name', 'remote_as', 'peer_ip',
                         'auth_type', 'password']
             res = dict((k, ri[k]) for k in res_keys)
@@ -242,17 +246,17 @@ class BgpDbMixin(common_db.CommonDbMixin):
 
     def get_bgp_peers(self, context, fields=None, filters=None, sorts=None,
                       limit=None, marker=None, page_reverse=False):
-        return self._get_collection(context, BgpPeer,
-                                    self._make_bgp_peer_dict,
-                                    filters=filters, fields=fields,
-                                    sorts=sorts, limit=limit,
-                                    page_reverse=page_reverse)
+        return model_query.get_collection(context, BgpPeer,
+                                          self._make_bgp_peer_dict,
+                                          filters=filters, fields=fields,
+                                          sorts=sorts, limit=limit,
+                                          page_reverse=page_reverse)
 
     def get_bgp_peers_by_bgp_speaker(self, context,
                                      bgp_speaker_id, fields=None):
         filters = [BgpSpeakerPeerBinding.bgp_speaker_id == bgp_speaker_id,
                    BgpSpeakerPeerBinding.bgp_peer_id == BgpPeer.id]
-        with db_api.context_manager.reader.using(context):
+        with db_api.CONTEXT_READER.using(context):
             query = context.session.query(BgpPeer)
             query = query.filter(*filters)
             return [self._make_bgp_peer_dict(x, fields) for x in query.all()]
@@ -262,13 +266,13 @@ class BgpDbMixin(common_db.CommonDbMixin):
         return self._make_bgp_peer_dict(bgp_peer_db, fields=fields)
 
     def delete_bgp_peer(self, context, bgp_peer_id):
-        with db_api.context_manager.writer.using(context):
+        with db_api.CONTEXT_WRITER.using(context):
             bgp_peer_db = self._get_bgp_peer(context, bgp_peer_id)
             context.session.delete(bgp_peer_db)
 
     def update_bgp_peer(self, context, bgp_peer_id, bgp_peer):
         bp = bgp_peer[bgp_ext.BGP_PEER_BODY_KEY_NAME]
-        with db_api.context_manager.writer.using(context):
+        with db_api.CONTEXT_WRITER.using(context):
             bgp_peer_db = self._get_bgp_peer(context, bgp_peer_id)
             if ((bp.get('password') is not None) and
                 (bgp_peer_db['auth_type'] == 'none')):
@@ -280,13 +284,13 @@ class BgpDbMixin(common_db.CommonDbMixin):
 
     def _get_bgp_speaker(self, context, bgp_speaker_id):
         try:
-            return self._get_by_id(context, BgpSpeaker,
-                                   bgp_speaker_id)
+            return model_query.get_by_id(context, BgpSpeaker,
+                                         bgp_speaker_id)
         except sa_exc.NoResultFound:
             raise bgp_ext.BgpSpeakerNotFound(id=bgp_speaker_id)
 
     def _get_bgp_speaker_ids_by_router(self, context, router_id):
-        with db_api.context_manager.reader.using(context):
+        with db_api.CONTEXT_READER.using(context):
             network_binding = aliased(BgpSpeakerNetworkBinding)
             r_port = aliased(l3_db.RouterPort)
             query = context.session.query(network_binding.bgp_speaker_id)
@@ -299,7 +303,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
             return [binding.bgp_speaker_id for binding in query.all()]
 
     def _get_bgp_speaker_ids_by_binding_network(self, context, network_id):
-        with db_api.context_manager.reader.using(context):
+        with db_api.CONTEXT_READER.using(context):
             query = context.session.query(
                                       BgpSpeakerNetworkBinding.bgp_speaker_id)
             query = query.filter(
@@ -322,7 +326,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
         return uuid
 
     def _get_bgp_peers_by_bgp_speaker_binding(self, context, bgp_speaker_id):
-        with db_api.context_manager.reader.using(context):
+        with db_api.CONTEXT_READER.using(context):
             query = context.session.query(BgpPeer)
             query = query.filter(
                      BgpSpeakerPeerBinding.bgp_speaker_id == bgp_speaker_id,
@@ -331,16 +335,16 @@ class BgpDbMixin(common_db.CommonDbMixin):
 
     def _save_bgp_speaker_peer_binding(self, context, bgp_speaker_id,
                                        bgp_peer_id):
-        with db_api.context_manager.writer.using(context):
+        with db_api.CONTEXT_WRITER.using(context):
             try:
-                bgp_speaker = self._get_by_id(context, BgpSpeaker,
-                                              bgp_speaker_id)
+                bgp_speaker = model_query.get_by_id(context, BgpSpeaker,
+                                                    bgp_speaker_id)
             except sa_exc.NoResultFound:
                 raise bgp_ext.BgpSpeakerNotFound(id=bgp_speaker_id)
 
             try:
-                bgp_peer = self._get_by_id(context, BgpPeer,
-                                           bgp_peer_id)
+                bgp_peer = model_query.get_by_id(context, BgpPeer,
+                                                 bgp_peer_id)
             except sa_exc.NoResultFound:
                 raise bgp_ext.BgpPeerNotFound(id=bgp_peer_id)
 
@@ -361,7 +365,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
 
     def _remove_bgp_speaker_peer_binding(self, context, bgp_speaker_id,
                                          bgp_peer_id):
-        with db_api.context_manager.writer.using(context):
+        with db_api.CONTEXT_WRITER.using(context):
 
             try:
                 binding = self._get_bgp_speaker_peer_binding(context,
@@ -377,16 +381,16 @@ class BgpDbMixin(common_db.CommonDbMixin):
                                           context,
                                           bgp_speaker_id,
                                           network_id):
-        with db_api.context_manager.writer.using(context):
+        with db_api.CONTEXT_WRITER.using(context):
             try:
-                bgp_speaker = self._get_by_id(context, BgpSpeaker,
-                                              bgp_speaker_id)
+                bgp_speaker = model_query.get_by_id(context, BgpSpeaker,
+                                                    bgp_speaker_id)
             except sa_exc.NoResultFound:
                 raise bgp_ext.BgpSpeakerNotFound(id=bgp_speaker_id)
 
             try:
-                network = self._get_by_id(context, models_v2.Network,
-                                          network_id)
+                network = model_query.get_by_id(context, models_v2.Network,
+                                                network_id)
             except sa_exc.NoResultFound:
                 raise n_exc.NetworkNotFound(net_id=network_id)
 
@@ -398,7 +402,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
 
     def _remove_bgp_speaker_network_binding(self, context,
                                             bgp_speaker_id, network_id):
-        with db_api.context_manager.writer.using(context):
+        with db_api.CONTEXT_WRITER.using(context):
 
             try:
                 binding = self._get_bgp_speaker_network_binding(
@@ -420,27 +424,27 @@ class BgpDbMixin(common_db.CommonDbMixin):
         res = dict((k, bgp_speaker[k]) for k in attrs)
         res['peers'] = [x.bgp_peer_id for x in peer_bindings]
         res['networks'] = [x.network_id for x in network_bindings]
-        return self._fields(res, fields)
+        return db_utils.resource_fields(res, fields)
 
     def _make_advertised_routes_dict(self, routes):
         return {'advertised_routes': list(routes)}
 
     def _get_bgp_peer(self, context, bgp_peer_id):
         try:
-            return self._get_by_id(context, BgpPeer, bgp_peer_id)
+            return model_query.get_by_id(context, BgpPeer, bgp_peer_id)
         except sa_exc.NoResultFound:
             raise bgp_ext.BgpPeerNotFound(id=bgp_peer_id)
 
     def _get_bgp_speaker_peer_binding(self, context,
                                       bgp_speaker_id, bgp_peer_id):
-        query = self._model_query(context, BgpSpeakerPeerBinding)
+        query = model_query.query_with_hooks(context, BgpSpeakerPeerBinding)
         return query.filter(
                         BgpSpeakerPeerBinding.bgp_speaker_id == bgp_speaker_id,
                         BgpSpeakerPeerBinding.bgp_peer_id == bgp_peer_id).one()
 
     def _get_bgp_speaker_network_binding(self, context,
                                          bgp_speaker_id, network_id):
-        query = self._model_query(context, BgpSpeakerNetworkBinding)
+        query = model_query.query_with_hooks(context, BgpSpeakerNetworkBinding)
         return query.filter(
                     BgpSpeakerNetworkBinding.bgp_speaker_id == bgp_speaker_id,
                     BgpSpeakerNetworkBinding.network_id == network_id).one()
@@ -449,10 +453,10 @@ class BgpDbMixin(common_db.CommonDbMixin):
         attrs = ['tenant_id', 'id', 'name', 'peer_ip', 'remote_as',
                  'auth_type', 'password']
         res = dict((k, bgp_peer[k]) for k in attrs)
-        return self._fields(res, fields)
+        return db_utils.resource_fields(res, fields)
 
     def _get_address_scope_ids_for_bgp_speaker(self, context, bgp_speaker_id):
-        with db_api.context_manager.reader.using(context):
+        with db_api.CONTEXT_READER.using(context):
             binding = aliased(BgpSpeakerNetworkBinding)
             address_scope = aliased(address_scope_db.AddressScope)
             query = context.session.query(address_scope)
@@ -466,7 +470,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
 
     def get_routes_by_bgp_speaker_id(self, context, bgp_speaker_id):
         """Get all routes that should be advertised by a BgpSpeaker."""
-        with db_api.context_manager.reader.using(context):
+        with db_api.CONTEXT_READER.using(context):
             net_routes = self._get_tenant_network_routes_by_bgp_speaker(
                                                                context,
                                                                bgp_speaker_id)
@@ -476,12 +480,16 @@ class BgpDbMixin(common_db.CommonDbMixin):
             dvr_fip_routes = self._get_dvr_fip_host_routes_by_bgp_speaker(
                                                                context,
                                                                bgp_speaker_id)
-            return itertools.chain(fip_routes, net_routes, dvr_fip_routes)
+            dvr_fixedip_routes = self._get_dvr_fixed_ip_routes_by_bgp_speaker(
+                                                            context,
+                                                            bgp_speaker_id)
+            return itertools.chain(fip_routes, net_routes, dvr_fip_routes,
+                                   dvr_fixedip_routes)
 
     def get_routes_by_bgp_speaker_binding(self, context,
                                           bgp_speaker_id, network_id):
         """Get all routes for the given bgp_speaker binding."""
-        with db_api.context_manager.reader.using(context):
+        with db_api.CONTEXT_READER.using(context):
             fip_routes = self._get_central_fip_host_routes_by_binding(
                                                                context,
                                                                network_id,
@@ -520,7 +528,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
     def _get_central_fip_host_routes_by_router(self, context, router_id,
                                                bgp_speaker_id):
         """Get floating IP host routes with the given router as nexthop."""
-        with db_api.context_manager.reader.using(context):
+        with db_api.CONTEXT_READER.using(context):
             dest_alias = aliased(l3_db.FloatingIP,
                                  name='destination')
             next_hop_alias = aliased(models_v2.IPAllocation,
@@ -555,7 +563,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
 
     def _get_dvr_fip_host_routes_by_router(self, context, bgp_speaker_id,
                                            router_id):
-        with db_api.context_manager.reader.using(context):
+        with db_api.CONTEXT_READER.using(context):
             gw_query = self._get_gateway_query(context, bgp_speaker_id)
 
             fip_query = self._get_fip_query(context, bgp_speaker_id)
@@ -571,7 +579,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
     def _get_central_fip_host_routes_by_binding(self, context,
                                                 network_id, bgp_speaker_id):
         """Get all floating IP host routes for the given network binding."""
-        with db_api.context_manager.reader.using(context):
+        with db_api.CONTEXT_READER.using(context):
             # Query the DB for floating IP's and the IP address of the
             # gateway port
             dest_alias = aliased(l3_db.FloatingIP,
@@ -609,7 +617,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
 
     def _get_dvr_fip_host_routes_by_binding(self, context, network_id,
                                             bgp_speaker_id):
-        with db_api.context_manager.reader.using(context):
+        with db_api.CONTEXT_READER.using(context):
             BgpBinding = BgpSpeakerNetworkBinding
 
             gw_query = self._get_gateway_query(context, bgp_speaker_id)
@@ -628,7 +636,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
     def _get_central_fip_host_routes_by_bgp_speaker(self, context,
                                                     bgp_speaker_id):
         """Get all the floating IP host routes advertised by a BgpSpeaker."""
-        with db_api.context_manager.reader.using(context):
+        with db_api.CONTEXT_READER.using(context):
             dest_alias = aliased(l3_db.FloatingIP,
                                  name='destination')
             next_hop_alias = aliased(models_v2.IPAllocation,
@@ -662,18 +670,21 @@ class BgpDbMixin(common_db.CommonDbMixin):
                  next_hop_alias.subnet_id == models_v2.Subnet.id,
                  models_v2.Subnet.ip_version == 4)
             query = query.outerjoin(router_attrs,
-                                    l3_db.Router.id == router_attrs.router_id)
+                                    router_attrs.router_id == l3_db.Router.id)
             query = query.filter(router_attrs.distributed != sa.sql.true())
             return self._host_route_list_from_tuples(query.all())
 
     def _get_gateway_query(self, context, bgp_speaker_id):
         BgpBinding = BgpSpeakerNetworkBinding
+        AddressScope = address_scope_db.AddressScope
         ML2PortBinding = ml2_models.PortBinding
         IpAllocation = models_v2.IPAllocation
         Port = models_v2.Port
-        gw_query = context.session.query(Port.network_id,
-                                         ML2PortBinding.host,
-                                         IpAllocation.ip_address)
+        gw_query = context.session.query(
+                                Port.network_id,
+                                ML2PortBinding.host,
+                                IpAllocation.ip_address,
+                                AddressScope.id.label('address_scope_id'))
 
         #Subquery for FIP agent gateway ports
         gw_query = gw_query.filter(
@@ -681,6 +692,9 @@ class BgpDbMixin(common_db.CommonDbMixin):
             IpAllocation.port_id == Port.id,
             IpAllocation.subnet_id == models_v2.Subnet.id,
             models_v2.Subnet.ip_version == 4,
+            models_v2.Subnet.subnetpool_id == models_v2.SubnetPool.id,
+            models_v2.SubnetPool.address_scope_id == AddressScope.id,
+            Port.network_id == models_v2.Subnet.network_id,
             Port.device_owner == lib_consts.DEVICE_OWNER_AGENT_GW,
             Port.network_id == BgpBinding.network_id,
             BgpBinding.bgp_speaker_id == bgp_speaker_id,
@@ -702,10 +716,53 @@ class BgpDbMixin(common_db.CommonDbMixin):
             BgpBinding.bgp_speaker_id == bgp_speaker_id)
         return fip_query
 
+    def _get_dvr_fixed_ip_query(self, context, bgp_speaker_id):
+        AddressScope = address_scope_db.AddressScope
+        ML2PortBinding = ml2_models.PortBinding
+        Port = models_v2.Port
+        IpAllocation = models_v2.IPAllocation
+
+        fixed_ip_query = context.session.query(
+            ML2PortBinding.host,
+            IpAllocation.ip_address,
+            IpAllocation.subnet_id,
+            AddressScope.id.label('address_scope_id'))
+        fixed_ip_query = fixed_ip_query.filter(
+            Port.id == ML2PortBinding.port_id,
+            IpAllocation.port_id == Port.id,
+            Port.device_owner.startswith(
+                lib_consts.DEVICE_OWNER_COMPUTE_PREFIX),
+            IpAllocation.subnet_id == models_v2.Subnet.id,
+            models_v2.Subnet.subnetpool_id == models_v2.SubnetPool.id,
+            AddressScope.id == models_v2.SubnetPool.address_scope_id)
+        return fixed_ip_query
+
+    def _get_dvr_fixed_ip_routes_by_bgp_speaker(self, context,
+                                                bgp_speaker_id):
+        with db_api.CONTEXT_READER.using(context):
+            gw_query = self._get_gateway_query(context, bgp_speaker_id)
+            fixed_query = self._get_dvr_fixed_ip_query(context,
+                                                       bgp_speaker_id)
+            join_query = self._join_fixed_by_host_binding_to_agent_gateway(
+                                                    context,
+                                                    fixed_query.subquery(),
+                                                    gw_query.subquery())
+            return self._host_route_list_from_tuples(join_query.all())
+
+    def _join_fixed_by_host_binding_to_agent_gateway(self, context,
+                                                   fixed_subq, gw_subq):
+        join_query = context.session.query(fixed_subq.c.ip_address,
+                                           gw_subq.c.ip_address)
+        and_cond = and_(
+                gw_subq.c.host == fixed_subq.c.host,
+                gw_subq.c.address_scope_id == fixed_subq.c.address_scope_id)
+
+        return join_query.join(gw_subq, and_cond)
+
     def _get_dvr_fip_host_routes_by_bgp_speaker(self, context,
                                                 bgp_speaker_id):
         router_attrs = l3_attrs_db.RouterExtraAttributes
-        with db_api.context_manager.reader.using(context):
+        with db_api.CONTEXT_READER.using(context):
             gw_query = self._get_gateway_query(context, bgp_speaker_id)
             fip_query = self._get_fip_query(context, bgp_speaker_id)
 
@@ -734,7 +791,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
                                               network_id, bgp_speaker_id):
         """Get all tenant network routes for the given network."""
 
-        with db_api.context_manager.reader.using(context):
+        with db_api.CONTEXT_READER.using(context):
             tenant_networks_query = self._tenant_networks_by_network_query(
                                                                context,
                                                                network_id,
@@ -753,7 +810,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
                                              bgp_speaker_id):
         """Get all tenant network routes with the given router as nexthop."""
 
-        with db_api.context_manager.reader.using(context):
+        with db_api.CONTEXT_READER.using(context):
             scopes = self._get_address_scope_ids_for_bgp_speaker(
                                                                context,
                                                                bgp_speaker_id)
@@ -800,7 +857,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
                                                   bgp_speaker_id):
         """Get all tenant network routes to be advertised by a BgpSpeaker."""
 
-        with db_api.context_manager.reader.using(context):
+        with db_api.CONTEXT_READER.using(context):
             tenant_nets_q = self._tenant_networks_by_bgp_speaker_query(
                                                                context,
                                                                bgp_speaker_id)
@@ -937,7 +994,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
            models_v2.IPAllocation.subnet_id == models_v2.Subnet.id]
 
     def _tenant_prefixes_by_router(self, context, router_id, bgp_speaker_id):
-        with db_api.context_manager.reader.using(context):
+        with db_api.CONTEXT_READER.using(context):
             query = context.session.query(models_v2.Subnet.cidr.distinct())
             filters = self._tenant_prefixes_by_router_filters(router_id,
                                                               bgp_speaker_id)
@@ -965,7 +1022,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
                                              context,
                                              router_port_id,
                                              bgp_speaker_id):
-        with db_api.context_manager.reader.using(context):
+        with db_api.CONTEXT_READER.using(context):
             query = context.session.query(models_v2.Subnet.cidr.distinct())
             filters = self._tenant_prefixes_by_router_filters(router_port_id,
                                                               bgp_speaker_id)
@@ -988,7 +1045,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
 
     def _bgp_speakers_for_gateway_network(self, context, network_id):
         """Return all BgpSpeakers for the given gateway network"""
-        with db_api.context_manager.reader.using(context):
+        with db_api.CONTEXT_READER.using(context):
             query = context.session.query(BgpSpeaker)
             query = query.filter(
                   BgpSpeakerNetworkBinding.network_id == network_id,
@@ -998,7 +1055,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
     def _bgp_speakers_for_gw_network_by_family(self, context,
                                                network_id, ip_version):
         """Return the BgpSpeaker by given gateway network and ip_version"""
-        with db_api.context_manager.reader.using(context):
+        with db_api.CONTEXT_READER.using(context):
             query = context.session.query(BgpSpeaker)
             query = query.filter(
                   BgpSpeakerNetworkBinding.network_id == network_id,
@@ -1023,7 +1080,7 @@ class BgpDbMixin(common_db.CommonDbMixin):
 
     def _get_router(self, context, router_id):
         try:
-            router = self._get_by_id(context, l3_db.Router, router_id)
+            router = model_query.get_by_id(context, l3_db.Router, router_id)
         except sa_exc.NoResultFound:
             raise l3_exc.RouterNotFound(router_id=router_id)
         return router
@@ -1083,3 +1140,61 @@ class BgpDbMixin(common_db.CommonDbMixin):
             return
         except sa_exc.MultipleResultsFound:
             return
+
+    def get_external_networks_for_port(self, ctx, port,
+                                       match_address_scopes=True):
+        with db_api.CONTEXT_READER.using(ctx):
+            # Retrieve address scope info for the supplied port
+            port_fixed_ips = port.get('fixed_ips')
+            if not port_fixed_ips:
+                return []
+            subnets_filter = {'id': [x['subnet_id'] for x in port_fixed_ips]}
+            port_subnets = subnet_obj.Subnet.get_objects(ctx, **subnets_filter)
+            port_subnetpools = subnetpool_obj.SubnetPool.get_objects(
+                        ctx, id=[x.subnetpool_id for x in port_subnets])
+            port_scopes = set([x.address_scope_id for x in port_subnetpools])
+            if match_address_scopes and len(port_scopes) == 0:
+                return []
+
+            # Get all router IDs with an interface on the given port's network
+            router_iface_filters = {'device_owner':
+                                    [DEVICE_OWNER_ROUTER_INTF,
+                                     DEVICE_OWNER_DVR_INTERFACE],
+                                    'network_id': port['network_id']}
+            router_ids = [x.device_id for x in ports.Port.get_objects(
+                                                ctx, **router_iface_filters)]
+
+            # Retrieve the gateway ports for the identified routers
+            gw_port_filters = {'device_owner': DEVICE_OWNER_ROUTER_GW,
+                               'device_id': router_ids}
+            gw_ports = ports.Port.get_objects(ctx, **gw_port_filters)
+
+            # If we don't need to match address scopes, return here
+            if not match_address_scopes:
+                return list(set([x.network_id for x in gw_ports]))
+
+            # Retrieve address scope info for associated gateway networks
+            gw_fixed_ips = []
+            for gw_port in gw_ports:
+                gw_fixed_ips.extend(gw_port.fixed_ips)
+            gw_subnet_filters = {'id': [x.subnet_id for x in gw_fixed_ips]}
+            gw_subnets = subnet_obj.Subnet.get_objects(ctx,
+                                                       **gw_subnet_filters)
+            ext_net_subnetpool_map = {}
+            for gw_subnet in gw_subnets:
+                ext_net_id = gw_subnet.network_id
+                ext_pool = subnetpool_obj.SubnetPool.get_object(
+                            ctx, id=gw_subnet.subnetpool_id)
+                ext_scope_set = ext_net_subnetpool_map.get(ext_net_id, set())
+                ext_scope_set.add(ext_pool.address_scope_id)
+                ext_net_subnetpool_map[ext_net_id] = ext_scope_set
+
+            ext_nets = []
+
+            # Match address scopes between port and gateway network(s)
+            for net in ext_net_subnetpool_map.keys():
+                ext_scopes = ext_net_subnetpool_map[net]
+                if ext_scopes.issubset(port_scopes):
+                    ext_nets.append(net)
+
+            return ext_nets
